// naumann@stce.rwth-aachen.de

#include <iostream> 
#include <cassert> 
#include <vector> 
#include <string> 
#include <fstream> 
#include <cmath> 
#include <limits>

struct plotable {
  virtual void plot() const=0;  
};

template<typename T>
class data : public plotable {
  std::vector<std::pair<T,T>> observations;
  std::string filename;
public:
  data(char*);
  ~data();
  void add_observation(const T&, const T&);
  const std::vector<std::pair<T,T>>& get_observations() const;
  void plot() const;
};


template<typename T>
data<T>::data(char* fn) : filename(fn) {
  std::ifstream ifs(filename); 
  size_t m=0; ifs >> m; 
  T x,y; 
  for (size_t i=0;i<m;i++) {
    ifs >> x >> y; observations.push_back(std::make_pair(x,y));
  }
  ifs.close(); 
}

template<typename T>
data<T>::~data() {
  std::ofstream ofs(filename); 
  ofs << observations.size() << std::endl; 
  for (const auto& o : observations) 
    ofs << o.first << " " << o.second << std::endl;
  ofs.close(); 
}

template<typename T>
void data<T>::add_observation(const T& x, const T &y) {
  observations.push_back(std::make_pair(x,y));
}

template<typename T>
const std::vector<std::pair<T,T>>& data<T>::get_observations() const {
  return observations;
}

template<typename T>
void data<T>::plot() const {
  std::ofstream ofs("data.plot"); 
  for (const auto& o : observations) 
    ofs << o.first << " " << o.second << std::endl;
  ofs.close(); 
}

template<typename T>
class model : public plotable{
  const data<T>& d;
  const T eps=std::numeric_limits<T>::epsilon();
  std::string filename;
  T mse() const; 
  T d_mse();
  T dd_mse();
protected:
  T p;
public:
  model(const data<T>&, char*);
  virtual ~model();
  virtual T evaluate(T) const=0;
  void calibrate();
  void plot() const;
};

template<typename T>
model<T>::model(const data<T>& d, char* filename) : d(d), filename(filename) {
  std::ifstream ifs(filename); 
  ifs >> p; 
  ifs.close(); 
}

template<typename T>
model<T>::~model() {
  std::ofstream ofs(filename); 
  ofs << p << std::endl;
  ofs.close(); 
}

template<typename T>
T model<T>::mse() const {
  T e=0;
  for (const auto& o : d.get_observations()) 
    e+=pow(evaluate(o.first)-o.second,2);
  return e;
}

template<typename T>
T model<T>::d_mse() {
  const T delta_p=(p!=0) ? sqrt(sqrt(eps))*fabs(p) : sqrt(sqrt(eps));
  p+=delta_p; T e=mse(); p-=delta_p;
  return (e-mse())/delta_p;
}

template<typename T>
T model<T>::dd_mse() {
  const T delta_p=(p!=0) ? sqrt(sqrt(eps))*fabs(p) : sqrt(sqrt(eps));
  p+=delta_p; T e=d_mse(); p-=delta_p;
  return (e-d_mse())/delta_p;
}

template<typename T>
void model<T>::calibrate() {
  while (fabs(d_mse())>sqrt(eps)) p-=d_mse()/dd_mse();
}

template<typename T>
void model<T>::plot() const {
  T max_x=0;
  for (const auto& o : d.get_observations()) 
    max_x=std::max(o.first,max_x);
  std::ofstream ofs("model.plot"); 
  for (T x=0;x<=max_x;x+=1e-1)
    ofs << x << " " << evaluate(x) << std::endl;
  ofs.close(); 
}

template<typename T>
struct linear_model : public model<T> {
  using model<T>::p;
  linear_model(const data<T>&, char*);
  ~linear_model();
  T evaluate(T) const;
};


template<typename T>
linear_model<T>::linear_model(const data<T> &d, char* filename) : model<T>(d,filename) {}

template<typename T>
linear_model<T>::~linear_model() {}

template<typename T>
T linear_model<T>::evaluate(T x) const { return p*x; }

template<typename T>
struct nonlinear_model : public model<T> {
  using model<T>::p;
  nonlinear_model(const data<T>&, char*);
  ~nonlinear_model();
  T evaluate(T) const;
};


template<typename T>
nonlinear_model<T>::nonlinear_model(const data<T> &d, char* filename) : model<T>(d,filename) {}

template<typename T>
nonlinear_model<T>::~nonlinear_model() {}

template<typename T>
T nonlinear_model<T>::evaluate(T x) const { return pow(p*x,2); }

int main(int argc, char *argv[]) {
  assert(argc==3); 
  using T=double;
  data<T> d(argv[1]); nonlinear_model<T> m(d,argv[2]); 
  T x,y;
  std::cout << "new observation:" << std::endl;
  std::cout << "x="; std::cin >> x;
  std::cout << "y="; std::cin >> y;
  d.add_observation(x,y);
  m.calibrate();
  std::cout << "new simulation:" << std::endl;
  std::cout << "x="; std::cin >> x;
  std::cout << "y=" << m.evaluate(x) << std::endl;
  d.plot(); m.plot();
  return 0;
}
